//- ****************************************************************************
//- 
//- Copyright 2009 Sandia Corporation. Under the terms of Contract 
//- DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains 
//- certain rights in this software.
//-
//- BSD Open Source License.
//- All rights reserved.
//- 
//- Redistribution and use in source and binary forms, with or without 
//- modification, are permitted provided that the following conditions are met:
//-
//-    * Redistributions of source code must retain the above copyright notice, 
//-      this list of conditions and the following disclaimer.
//-    * Redistributions in binary form must reproduce the above copyright 
//-      notice, this list of conditions and the following disclaimer in the 
//-      documentation and/or other materials provided with the distribution.
//-    * Neither the name of Sandia National Laboratories nor the names of its 
//-      contributors may be used to endorse or promote products derived from  
//-      this software without specific prior written permission.
//-
//- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//- ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//- POSSIBILITY OF SUCH DAMAGE.
//-


//- ****************************************************************************
//-
//- Program:       Uncertainty
//- Module:        $RCSfile: Uncertainty.h,v $
//- Revision:      $Revision: 1.16 $
//- Last Modified: $Date: 2011/08/23 20:56:55 $
//- Last Check-in: $Author: sballar $
//-
//- ****************************************************************************
#ifndef Uncertainty_H
#define Uncertainty_H

// **** _SYSTEM INCLUDES_ ******************************************************

#include <string>
#include <vector>

#include "SLBMGlobals.h"
#include "DataBuffer.h"

using namespace std;

// **** _BEGIN SLBM NAMESPACE_ **************************************************

namespace slbm {

// **** _LOCAL INCLUDES_ *******************************************************

//!
//! \brief A Uncertainty object contains the raw data to calculate a modeling
//! error in seconds as a function of distance in radians.
//!
//! A Uncertainty object contains the raw data to calculate a modeling
//! error in seconds as a function of distance in radians.
//!
class SLBM_EXP Uncertainty
{

public:

	//! \brief Default constructor.
	//!
	//! Default constructor.
	Uncertainty();

  // \brief Parameterized constructor that loads model error
	//! from a specified file.
	//!
	//! Parameterized constructor that loads model error data
	//! from a file. Uses the input phase string to find the uncertainty
  //! data file and assigns a phase ordering index.
  Uncertainty(string modelPath, const string& phase, int phasenum);

	//! \brief Parameterized constructor that loads model error
	//! from a specified file.
	//!
	//! Parameterized constructor that loads model error data
	//! from a file. Uses an SLBM specific phase ordering index.
	Uncertainty(string modelPath, int phase, int attribute);

	//! \brief Parameterized constructor that loads uncertainty data
	//! from the input DataBuffer.
	//!
	//! Parameterized constructor that loads model error data
	//! from a file.
  Uncertainty(util::DataBuffer& buffer);

	//! \brief Copy constructor.
	//!
	//! Copy constructor.
  Uncertainty(const Uncertainty& u);

	//! \brief Destructor.
	//!
	//! Destructor.
	~Uncertainty();

	//! \brief Assignment operator.
	//!
	//! Assignment operator.
  Uncertainty& operator=(const Uncertainty& u);

	//! \brief A public convenience accessor used to verify the error data for the 
	//! correct model phase is loaded in memory.
	//!
	//! A public convenience accessor used to verify the error data for the 
	//! correct model phase is loaded in memory.
	int getPhase();

	//! \brief A public convenience accessor used to verify the error data for the 
	//! correct model phase is loaded in memory.
	//!
	//! A public convenience accessor used to verify the error data for the 
	//! correct model phase is loaded in memory.
  const string& getPhaseStr();

	//! \brief Returns the model uncertainty as a function of angular distance 
	//! (radians) and depth.
	//!
	//! Returns the model uncertainty as a function of angular distance 
	//! (radians) and depth. Depth defaults to the surface (0.0).
	double getUncertainty(const double& distance, double depth = 0.0);

	//! \brief Returns the model variance as a function of angular distance 
	//! (radians) and depth.
	//!
	//! Returns the model variance as a function of angular distance 
	//! (radians) and depth. Depth defaults to the surface (0.0).
	double getVariance(const double& distance, double depth = 0.0);

	//! \brief Returns the model uncertainty DataBuffer size storage requirement.
	//!
	//! Returns the model uncertainty DataBuffer size storage requirement.
  int getBufferSize() const;

	//! \brief Writes the uncertainty object into the input DataBuffer.
	//!
	//! Writes the uncertainty object into the input DataBuffer.
  void serialize(util::DataBuffer& buffer);

  //! \brief Reads the uncertainty object from the input DataBuffer.
	//!
	//! Reads the uncertainty object from the input DataBuffer.
  void deserialize(util::DataBuffer& buffer);

private:

  //! \brief Returns the distance interpolated uncertainty at depth,
  //! \em idepth, and fraction \em f between distance indices \em idist
  //! and idist+1.
  //!
  //! Returns the distance interpolated uncertainty at depth,
  //! \em idepth, and fraction \em f between distance indices \em idist
  //! and \em idist+1. For \em f = 0 errVal[idepth][idist] is returned.
  //! For \em f = 1 errVal[idepth][idist+1] is returned
  double getUncertainty(double f, int idist, int idepth);

  //! \brief Returns the distance interpolated variance at depth,
  //! \em idepth, and fraction \em f between distance indices \em idist
  //! and idist+1.
  //!
  //! Returns the distance interpolated variance at depth,
  //! \em idepth, and fraction \em f between distance indices \em idist
  //! and \em idist+1. For \em f = 0 errVal[idepth][idist]^2 is returned.
  //! For \em f = 1 errVal[idepth][idist+1]^2 is returned
  double getVariance(double f, int idist, int idepth);

  //! \brief Private function used to find the bracketing index for performing
  //! uncertainty / variance distance interpolation.
  //!
  //! Private function used to find the bracketing index for performing
  //! uncertainty / variance distance interpolation. The bracketing \em index
  //! is found in vector \em v such that
  //
  //!    \em v[index] <= \em x <= \em v[index+1].
  //
  //! The weight \em w is calculated as
  //
  //!    \em w = (x - v[index]) / (v[index+1] - v[index]);
  void getIndex(double x, const vector<double>& v, int& index, double& w);

	//! \brief A function called by the constructor to read model error data
	//! from an ASCII text file.
	//!
	//! \brief A function called by the constructor to read model error data
	//! from an ASCII text file.
	void readFile(const string& filename);

	//! \brief The seismic phase for which modeling error data was loaded.
	//!
	//! The seismic phase for which modeling error data was loaded.
	int phaseNum;

	//! \brief The string version of the seismic phase for which modeling error
	//! data was loaded.
	//!
	//! The string version of the seismic phase for which modeling error
	//! data was loaded.
	string phaseStr;

	//! \brief The attribute (TT, Sh or Az) for which modeling error
	//! data was loaded.
	//!
	//! The attribute (TT, Sh or Az) for which modeling error
	//! data was loaded.
	int attributeNum;

	string attributeStr;

	//! \brief The path and filename for the currently loaded seismic phase
	//! modeling error file.
	//!
	//! The path and filename for the currently loaded seismic phase
	//! modeling error file.
	string fname;

	//! \brief A vector of doubles representing the angular distances (in degrees)
	// for which the model errors are defined.
	//!
	//! A vector of doubles representing the angular distances (in degrees)
	// for which the model errors are defined.
	vector<double> errDistances;

	//! \brief A vector of doubles representing the depths
	// for which the model errors are defined.
	//!
	//! A vector of doubles representing the depths
	// for which the model errors are defined.
	vector<double> errDepths;

	//! \brief A vector of doubles representing the model errors (in seconds)
	// corresponding to the defined depths (first index) and angular distances.
	//!
	//! A vector of doubles representing the model errors (in seconds)
	// corresponding to the defined depths (first index) and angular distances.
	vector<vector<double> > errVal;

};

inline int Uncertainty::getPhase()
{
	return phaseNum;
}

inline const string& Uncertainty::getPhaseStr()
{
	return phaseStr;
}

inline double Uncertainty::getUncertainty(double f, int idist, int idepth)
{
	return (f * (errVal[idepth][idist+1] - errVal[idepth][idist]) +
          errVal[idepth][idist]);
}

inline double Uncertainty::getVariance(double f, int idist, int idepth)
{
	return (f * (errVal[idepth][idist+1] * errVal[idepth][idist+1] -
               errVal[idepth][idist] * errVal[idepth][idist]) +
              errVal[idepth][idist] * errVal[idepth][idist]);
}

} // end slbm namespace

#endif // Uncertainty_H
