//- ****************************************************************************
//- 
//- Copyright 2009 Sandia Corporation. Under the terms of Contract 
//- DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains 
//- certain rights in this software.
//-
//- BSD Open Source License.
//- All rights reserved.
//- 
//- Redistribution and use in source and binary forms, with or without 
//- modification, are permitted provided that the following conditions are met:
//-
//-    * Redistributions of source code must retain the above copyright notice, 
//-      this list of conditions and the following disclaimer.
//-    * Redistributions in binary form must reproduce the above copyright 
//-      notice, this list of conditions and the following disclaimer in the 
//-      documentation and/or other materials provided with the distribution.
//-    * Neither the name of Sandia National Laboratories nor the names of its 
//-      contributors may be used to endorse or promote products derived from  
//-      this software without specific prior written permission.
//-
//- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//- ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//- POSSIBILITY OF SUCH DAMAGE.
//-


//- ****************************************************************************
//-
//- Program:       Grid
//- Module:        $RCSfile: Grid.cc,v $
//- Revision:      $Revision: 1.48 $
//- Last Modified: $Date: 2011/11/30 22:41:28 $
//- Last Check-in: $Author: sballar $
//-
//- ****************************************************************************
#include "Grid.h"
#include "SLBMGlobals.h"

#include <sys/stat.h> 
#include <errno.h>

#if defined(_WIN32) || defined(WIN32) 
  // Windows compiler
  #include <direct.h>       // Windows directory creation/removal 
  #include <io.h>           // Windows file/directory existence checking
  #define PGL_MKDIR_OPTIONS // Windows - no options
#else                       
  // Sun compiler
  #include <dirent.h>       // Sun directory creation/removal
  #define PGL_MKDIR_OPTIONS , S_IRWXU | S_IRWXG | S_IRWXO  // Sun mkdir param 2
#endif

#include <string>
#include <vector>
#include <iostream>
#include <fstream>

//using namespace std;

// **** _BEGIN SLBM NAMESPACE_ **************************************************

namespace slbm {

// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Parameterized Grid Constructor
//
// *****************************************************************************
Grid::Grid()
: 	tessId(""),
	outputDirectory("")
{
	sources = new CrustalProfileStore(*this, 10);
	receivers = new CrustalProfileStore(*this, 1000);
}  // END Grid Default Constructor


// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Grid Destructor
//
// *****************************************************************************
Grid::~Grid()
{
	clear();
	delete sources;
	delete receivers;
}  // END RayTracer Destructor

bool Grid::operator==(const Grid& other)
{
	if (profiles.size() != other.profiles.size()) return false;

	for (int i=0; i< (int)profiles.size(); i++)
			if (profiles[i] != other.profiles[i]) 
				return false;

    return true;
}

void Grid::clear()
{
	clearCrustalProfiles();

	for (unsigned i=0; i<profiles.size(); i++)
		if (profiles[i]) delete profiles[i];
	profiles.clear();

	for (unsigned i=0; i<triangles.size(); i++)
		if (triangles[i]) delete triangles[i];
	triangles.clear();

	specialTriangles.clear();

	for (unsigned i=0; i<geoStacks.size(); i++)
		if (geoStacks[i]) delete geoStacks[i];
	geoStacks.clear();

	specialTriangles.clear();

	activeNodes.clear();
}

size_t Grid::memSize()  
{ 
	return 
		  sizeof(*geoStacks[0])*geoStacks.size()
		+ sizeof(*profiles[0])*profiles.size()
		+ sizeof(*triangles[0])*triangles.size()
		+ sizeof(geoStacks)
		+ sizeof(profiles)
		+ sizeof(triangles)
		+ sizeof(specialTriangles)
		+ sizeof(modelFileName)
		+ sizeof(cos_min)
		+ sizeof(V0)
		+ receivers->memSize()
		+ sources->memSize()
		;
}

void Grid::loadVelocityModel(const string& filename)
{
	// save the name of the file from which the velocity model
	// is being loaded.
	modelFileName = filename.c_str();

	ifstream fin;

	//open the file.  Return an error if cannot open the file.
	fin.open(filename.c_str());
	if (fin.fail() || !fin.is_open())
	{
		ostringstream os;
		os << endl << "ERROR in Grid::loadVelocityModel" << endl
			<<"Could not open file " << filename << endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),101);
	}

	int nStacks, nNodes, nTriangles, nData;
	fin >> nStacks >> nNodes >> nTriangles >> nData;

	if (nData != 24)
	{
		ostringstream os;
		os << endl << "ERROR in Grid::loadVelocityModel" << endl
			<<"File " << filename << " specifies a model with " << nData << endl
			<<" data items per layer but the SLBM code expects models with 24 data items per layers." 
			<< endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),503);
	}

	// read the average P and S velocities of the top of the mantle.  These
	// are averages over the part of the model that come from the Unified Model only.
	fin >> V0[PWAVE] >> V0[SWAVE];

	double lat, lon, elev, waterThick, depth[NLAYERS], vp[NLAYERS], vs[NLAYERS], g[2];
	int stackId;

	geoStacks.clear();
	geoStacks.resize(nStacks);

	for (int i=0; i<nStacks; i++)
	{
		for (int k=0; k<NLAYERS; k++)
		{
			if (k == 0)
			{
				// this is the water layer
				depth[k] = 0.;
				vp[k] = 1.5;
				vs[k] = 0.;
			}
			else
			{
				if (k == 1 || k == 6)
					depth[k] = depth[k-1];
				else 
					fin >> depth[k];
				fin >> vp[k] >> vs[k];
			}
		}
		fin >> g[0] >> g[1];

		geoStacks[i] = new GeoStack(i, depth, vp, vs, g);
	}

	profiles.clear();
	profiles.resize(nNodes);
	for (int i=0; i<nNodes; i++)
	{
		fin >> lat >> lon >> elev >> waterThick >> stackId;
		lat *= DEG_TO_RAD;
		lon *= DEG_TO_RAD;
		profiles[i] = new GridProfile(*this, i, lat, lon, elev, 
			abs(waterThick), geoStacks[stackId]);
		geoStacks[stackId]->incRefCount();
	}

	// create a temporary storage to hold a list of the triangles that 
	// each node is a member of.  There will a vector<Triangle*> for 
	// each node.  Each vector will contain pointers to the triangles
	// of which that node is a member.  There will almost always be 6 
	// of these, but in 12 cases there will only be 5.
	vector< vector<Triangle*> > triset(profiles.size());
	for (int i=0; i<triset.size(); i++)
		triset[i].reserve(6);

	triangles.resize(nTriangles);
	int a, b, c;
	for (int i=0; i<nTriangles; i++)
	{
		fin >> a >> b >> c;
		triangles[i] = new Triangle(i, profiles[a], profiles[b], profiles[c]);

		triset[a].push_back(triangles[i]);
		triset[b].push_back(triangles[i]);
		triset[c].push_back(triangles[i]);
	}

	// establish adjacency information for the triangles in the tessellation.
  defineTessAdjacency(nNodes, triset);

	fin.close();
	activeNodes.clear();

	tessId = "";
}

void Grid::saveVelocityModel(const string& filename)
{
	if (filename == modelFileName)
	{
		ostringstream os;
		os << endl << "ERROR in Grid::saveVelocityModel" << endl
			<<"Output file name cannot equal input file name."<< endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),102);
	}

	ofstream fout;

	//open the file.  Return an error if cannot open the file.
	fout.open(filename.c_str());
	if (fout.fail() || !fout.is_open())
	{
		ostringstream os;
		os << endl << "ERROR in Grid::saveVelocityModel" << endl
			<<"Could not open file " << filename << endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),103);
	}

	//fout << setiosflags(ios::fixed) << setiosflags(ios::showpoint);

	fout << geoStacks.size() 
		<< "  " << profiles.size() 
		<< " " << triangles.size()
		<< " 24" << endl;

	fout  << setprecision(7)
		<< V0[PWAVE] << " " << V0[SWAVE] << endl;

	for (int i=0; i<geoStacks.size(); i++)
	{
		for (int k=1; k<NLAYERS; k++)
		{
			if (k != 1 & k != 6)
				fout  << " " << geoStacks[i]->getDepth(k);

			fout << " " << geoStacks[i]->getVelocity(PWAVE, k) 
				 << " " << geoStacks[i]->getVelocity(SWAVE, k);
		}

		fout << " " << geoStacks[i]->getMantleGradient(PWAVE)
			<< " " << geoStacks[i]->getMantleGradient(SWAVE)
			<< endl;
	}

	for (int i=0; i<profiles.size(); i++)
		fout  << setprecision(10)
			<< profiles[i]->getLat() * RAD_TO_DEG << " " 
			<< profiles[i]->getLon() * RAD_TO_DEG << " " 
			<< setprecision(7)
			<< -profiles[i]->getDepth() << " " 
			<< profiles[i]->getWaterThick() << " " 
			<< profiles[i]->getGeoStackId()
			<< endl;

	for (int i=0; i<triangles.size(); i++)
		fout << triangles[i]->getNode(0)->getNodeId()
			<< " " << triangles[i]->getNode(1)->getNodeId()
			<< " " << triangles[i]->getNode(2)->getNodeId()
			<< endl;

	fout.close();
}

void Grid::loadVelocityModelBinary(util::DataBuffer& buffer)
{
	// ******************************************************
	// read geostack information
	// ******************************************************

  readGeoStacks(buffer);

	// ******************************************************
	// read connectivity information
	// ******************************************************

  vector<float> elev, waterThick;
	vector<int> stackId;
  int nNodes = 0;
  readConnectivity(buffer, nNodes, elev, waterThick, stackId);

	// ******************************************************
	// read Tessellation information
	// ******************************************************

	vector< vector<Triangle*> > triset;
  readTessellationData(buffer, nNodes, elev, waterThick,
	                     stackId, triset);

  // read and ignore metadata count

	buffer.readInt32();

	// establish adjacency information for the triangles in the tessellation.

  defineTessAdjacency(nNodes, triset);
}

void Grid::loadVelocityModelBinary(const string& dirname)
{
  util::DataBuffer buffer;
	string filename;

	clear();
	inputDirectory = dirname;

	// ******************************************************
	// read GeoStack file
	// ******************************************************

  filename = "geostacks";

	reaDataBuffererFromFile(buffer, dirname, filename);

  readGeoStacks(buffer);

	buffer.clear();

	// ******************************************************
	// read connectivity file
	// ******************************************************

	filename = "connectivity";

	reaDataBuffererFromFile(buffer, dirname, filename);

  vector<float> elev, waterThick;
	vector<int> stackId;
  int nNodes = 0;
  readConnectivity(buffer, nNodes, elev, waterThick, stackId);

	buffer.clear();

	// ******************************************************
	// read Tessellation file
	// ******************************************************

	filename = "../tess/"+tessId;

	reaDataBuffererFromFile(buffer, dirname, filename);

	vector< vector<Triangle*> > triset;
  readTessellationData(buffer, nNodes, elev, waterThick,
	                     stackId, triset);

	//cout << endl << "Metadata length: " << buffer->readInt32() << endl;

	// establish adjacency information for the triangles in the tessellation.
  defineTessAdjacency(nNodes, triset);
}

void Grid::readGeoStacks(util::DataBuffer& buffer)
{
  int nStacks = buffer.readInt32();
  //cout << endl << "GeoStack Count: " << nStacks << endl;

	// read the average P and S velocities of the top of the mantle.  These
	// are averages over the part of the model that come from the Unified Model only.
	V0[PWAVE] = buffer.readFloat();
	V0[SWAVE] = buffer.readFloat();

	//cout << endl << "Mantle velocities: " << V0[PWAVE] << ", " 
	//	<< V0[SWAVE] << endl;

	double depth[NLAYERS], vp[NLAYERS], vs[NLAYERS], g[2];

	geoStacks.resize(nStacks);

	for (int i=0; i<nStacks; i++)
	{
		for (int k=0; k<NLAYERS; k++)
		{
			if (k == 0)
			{
				// this is the water layer
				depth[k] = 0.;
				vp[k] = 1.5;
				vs[k] = 0.;
			}
			else
			{
				if (k == 6)
					// this the Pg/Lg velocity of middle crust
					depth[k] = depth[k-1];
				else 
					depth[k] = buffer.readFloat();
				vp[k] = buffer.readFloat();
				vs[k] = buffer.readFloat();
			}
		}
		g[0] = buffer.readFloat();
		g[1] = buffer.readFloat();

		geoStacks[i] = new GeoStack(i, depth, vp, vs, g);

	}
}

void Grid::readConnectivity(util::DataBuffer& buffer, int& nNodes,
                            vector<float>& elev, vector<float>& waterThick,
                            vector<int>& stackId)
{
	tessId = buffer.readString();
	//cout << "tess id: " << tessId << endl;

	// Read number of pairs (points)
	nNodes = buffer.readInt32();
	//cout << "Node Count: " << nNodes << endl;

	elev.reserve(nNodes);
	waterThick.reserve(nNodes);
	stackId.reserve(nNodes);
	for (int i=0; i<nNodes; i++)
	{
		stackId.push_back(buffer.readInt32());
		elev.push_back(buffer.readFloat());
		waterThick.push_back(buffer.readFloat());
	}
}

void Grid::readTessellationData(util::DataBuffer& buffer, int nNodes,
	                              const vector<float>& elev,
                                const vector<float>& waterThick,
	                              const vector<int>& stackId,
                                vector< vector<Triangle*> >& triset)
{
	// ******************************************************
	// read Tessellation file
	// ******************************************************

  // read header info
	string description = buffer.readString();
  //cout << "Description: " << description << endl;

	string parameters = buffer.readString();
  //cout << "Parameters: " << parameters << endl;
    
	string comments = buffer.readString();
  //cout << "Comments: " << comments << endl;
    
	// Read number of pairs (points)
  nNodes = buffer.readInt32();

  //cout << "Node Count: " << nNodes << endl;

	profiles.resize(nNodes);
	double lat, lon;
	
  for(int i = 0; i < nNodes; i++)
  {    
    lat = buffer.readDouble();
    lon = buffer.readDouble();
    profiles[i] = new GridProfile(*this, i, 
    lat * DEG_TO_RAD, // read latitude
    lon * DEG_TO_RAD, // read longitude
    (double)elev[i], 
    (double)abs(waterThick[i]), 
    geoStacks[stackId[i]]);

    geoStacks[stackId[i]]->incRefCount();

		//! throw an error if (1) any finite thickness layer above the middle crust
		//! has a velocity that exceeds the Pg/Lg velocity of the middle crust
		//! or (2) any finite thickness crustal layer has a velocity that exceeds
		//! the mantle velocity.
		if (geoStacks[stackId[i]]->hasLowVelocityZone())
		{
			ostringstream os;
			os << endl << "ERROR in Grid::loadVelocityModelBinary" << endl
				<< "Geostack has low velocity zone."  << endl 
				<< "Node ID, Stack Id = " << i << ", " << stackId[i] << endl
				<< "Node lat, lon = " << lat << ", " << lon << endl
				<< geoStacks[stackId[i]]->toString() << endl
				<< "Version " << SlbmVersion << "  File " << __FILE__ 
				<< " line " << __LINE__ << endl	<< endl;

			throw SLBMException(os.str(),504);
		}
	}

  int nTriangles = buffer.readInt32();
  //cout << endl << "Triangle Count: " << nTriangles << endl;

	// create a temporary storage to hold a list of the triangles that 
	// each node is a member of.  There will a vector<Triangle*> for 
	// each node.  Each vector will contain pointers to the triangles
	// of which that node is a member.  There will almost always be 6 
	// of these, but in 12 cases there will only be 5.
	//vector< vector<Triangle*> > triset(profiles.size());
  triset.resize(profiles.size());
	for (int i=0; i<triset.size(); i++)
		triset[i].reserve(6);

	triangles.resize(nTriangles);
	int a, b, c;
	for (int i=0; i<nTriangles; i++)
	{
		a = buffer.readInt32();
		b = buffer.readInt32();
		c = buffer.readInt32();
		triangles[i] = new Triangle(i, profiles[a], profiles[b], profiles[c]);

		triset[a].push_back(triangles[i]);
		triset[b].push_back(triangles[i]);
		triset[c].push_back(triangles[i]);
	}
}

void Grid::defineTessAdjacency(int nNodes,
                               const vector< vector<Triangle*> >& triset)
{
	// establish adjacency information for the triangles in the tessellation.
	bool found;
	GridProfile* nextNode;
	for (int i=0; i<nNodes; i++)
	{
		// for each node, iterate over the triangles of which it is a member.
		for (int j=0; j<triset[i].size(); j++)
		{
			// search for the index of this node in the current triangle 
			for (int k=0; k<3; k++)
			{
				if (triset[i][j]->getNode(k)->getNodeId() == profiles[i]->getNodeId())
				{
					// if the neighboring triangle for this edge has not yet been set
					if (triset[i][j]->getNeighbor(k) == NULL)
					{
						// identify the next node in the current triangle, moving 
						// clockwise from profiles[i];
						nextNode = triset[i][j]->getNode((k+1)%3);
						found = false;
						// iterate over the other triangles of which profiles[i] is a member
						for (int m=0; m<triset[i].size(); m++) if (m != j)
						{
							// search this triangle to see if it shares two nodes with the 
							// original triangle.
							for (int n=0; n<3; n++)
								if (triset[i][m]->getNode(n) == nextNode)
								{
									// these two triangles share two nodes.  Make them
									// neighbors.
									triset[i][j]->setNeighbor(k, triset[i][m]);
									triset[i][m]->setNeighbor(n, triset[i][j]);
									found = true;
									break;
								}
							if (found) break;
						}
					}
					break;
				}
			}
		}
	}

	// define the special triangles.  after a successful triangle walk,
  // the first special triangle will be set to the triangle just found
	// so that the next triangle search will check that triangle first.  
	// In addition, find the triangles which contain the first 42 nodes
	// in the model input file.  It is assumed that these nodes are 
	// evenly distributed over the surface of the earth.  Those triangles
	// will be saved as special triangles.  When Grid::findTriangle() is
	// called, the triangles in special triangles will be searched, in order,
	// for one whose node[0] is within 16 degrees of the location being searched
	// for.  If one is found, then the search will start there.  If none is found,
	// the walk will start at the closest one.
	findSpecialTriangles();
}

void Grid::specifyOutputDirectory(const string& dirname)
{
  util::DataBuffer buffer;
	outputDirectory = "";

	if (!is_directory(dirname))
		mkdir(dirname.c_str() PGL_MKDIR_OPTIONS);

  
	string s = "delete me";
	buffer.writeString(s);

	writeBufferToFile(buffer, dirname+"/deleteme.buf");

	remove((dirname+"/deleteme.buf").c_str());

	outputDirectory = dirname;
}

void Grid::saveVelocityModelBinary()
{
  util::DataBuffer buffer;
	string filename;

	if (outputDirectory == "")
	{
		ostringstream os;
		os << endl << "ERROR in Grid::saveVelocityModelBinary()" << endl
			<<"outputDirectory = <emptyString>.  Specify outputDirectory with call to method specifyOutpuDirectory()"<< endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),104);
	}
  
    //***NOTE reverse byte order on little endian machines
	if(!util::MD50::isBigEndian())
          buffer.setByteOrderReverse(true);

	if (tessId.length() == 0)
	{
		//******************************************************
		//write Tessellation file
		//******************************************************

		buffer.reserve(profiles.size()*2*sizeof(double)
			+triangles.size()*3*sizeof(int) + 1000);

		string s = "SLBM Tessellation";
		buffer.writeString(s.c_str());

		s = "Parameter list";
		buffer.writeString(s.c_str());

		s = "Comment";
		buffer.writeString(s.c_str());

		buffer.writeInt32(profiles.size());

		for (int i=0; i<profiles.size(); i++)
		{
			buffer.writeDouble(profiles[i]->getLatDegrees());
			buffer.writeDouble(profiles[i]->getLonDegrees());
		}

		buffer.writeInt32(triangles.size());

		for (int i=0; i<triangles.size(); i++)
			for (int j=0; j<3; j++)
				buffer.writeInt32(triangles[i]->getNode(j)->getNodeId());

		buffer.writeInt32(0); // no additional metadata

		tessId = buffer.generateDataBufMD5HashKey();

		filename = outputDirectory+"/../tess/"+tessId;

		writeBufferToFile(buffer, filename);

		buffer.clear();

		 //******************************************************
		 //write node adjacency file
		 //******************************************************
		buffer.reserve(profiles.size()*7*sizeof(int));

		buffer.writeString(tessId);

		buffer.writeInt32(profiles.size());

		vector<int> neighbors(profiles.size());
		vector<double> azimuth(profiles.size()), distance(profiles.size());

		map<double, int> nodeOrder;
		map<double, int>::iterator it;

		for (int i=0; i<profiles.size(); ++i)
		{
			getNodeNeighborInfo(profiles[i]->getNodeId(), neighbors, distance, azimuth);
			nodeOrder.clear();
			for (int j=0; j<neighbors.size(); ++j)
				nodeOrder[-azimuth[j]] = neighbors[j];

			buffer.writeInt32(neighbors.size());
			for (it = nodeOrder.begin(); it != nodeOrder.end(); ++it)
				buffer.writeInt32(it->second);
		}

		filename = outputDirectory+"/../tess/"+tessId+"_adjacency";

		writeBufferToFile(buffer, filename);

		buffer.clear();
	}

	// ******************************************************
	// write GeoStack file
	// ******************************************************
	buffer.reserve(geoStacks.size()*3*NLAYERS*4 + 1000);

    buffer.writeInt32(geoStacks.size());

	buffer.writeFloat((float)V0[PWAVE]);
	buffer.writeFloat((float)V0[SWAVE]);
	
	for (int i=0; i<geoStacks.size(); i++)
	{
		for (int k=1; k<NLAYERS; k++)
		{
			if (k != 6)
				buffer.writeFloat((float)geoStacks[i]->getDepth(k));
			buffer.writeFloat((float)geoStacks[i]->getVelocity(PWAVE, k));
			buffer.writeFloat((float)geoStacks[i]->getVelocity(SWAVE, k));
		}
		buffer.writeFloat((float)geoStacks[i]->getMantleGradient(PWAVE));
		buffer.writeFloat((float)geoStacks[i]->getMantleGradient(SWAVE));
	}

	filename = outputDirectory+"/geostacks";
	
	writeBufferToFile(buffer, filename);

	buffer.clear();

	// ******************************************************
	// write Connectivity file
	// ******************************************************
	buffer.reserve(profiles.size()*12 + 1000);

	buffer.writeString(tessId);

	buffer.writeInt32(profiles.size());

	for (int i=0; i<profiles.size(); i++)
	{
		buffer.writeInt32(profiles[i]->getGeoStackId());
		buffer.writeFloat((float)(-profiles[i]->getDepth()));
		buffer.writeFloat((float)profiles[i]->getWaterThick());
	}

	filename = outputDirectory+"/connectivity";

	writeBufferToFile(buffer, filename);

	buffer.clear();

}

int  Grid::getBufferSize() const
{
  int bsiz = 0;

  bsiz += 2*sizeof(int) + 2*sizeof(float);
  bsiz += 2*geoStacks.size()*sizeof(float);
  bsiz += 3*geoStacks.size()*(NLAYERS-1)*sizeof(float);
  bsiz += tessId.size() + 4*sizeof(int);
  bsiz += profiles.size()*(sizeof(int) + 2*sizeof(float));
  bsiz += 40+8*sizeof(int);
  bsiz += (profiles.size()*2 + 1)*sizeof(double);
  bsiz += 3*(triangles.size() + 1)*sizeof(int);

  return bsiz;
}

void Grid::saveVelocityModelBinary(util::DataBuffer& buffer)
{
	// ******************************************************
	// write GeoStack file
	// ******************************************************

  buffer.writeInt32(geoStacks.size());
	buffer.writeFloat((float)V0[PWAVE]);
	buffer.writeFloat((float)V0[SWAVE]);
	
	for (int i=0; i<geoStacks.size(); i++)
	{
		for (int k=1; k<NLAYERS; k++)
		{
			if (k != 6) buffer.writeFloat((float)geoStacks[i]->getDepth(k));
			buffer.writeFloat((float)geoStacks[i]->getVelocity(PWAVE, k));
			buffer.writeFloat((float)geoStacks[i]->getVelocity(SWAVE, k));
		}
		buffer.writeFloat((float)geoStacks[i]->getMantleGradient(PWAVE));
		buffer.writeFloat((float)geoStacks[i]->getMantleGradient(SWAVE));
	}

	// ******************************************************
	// write Connectivity file
	// ******************************************************

	buffer.writeString(tessId);
	buffer.writeInt32(profiles.size());

	for (int i=0; i<profiles.size(); i++)
	{
		buffer.writeInt32(profiles[i]->getGeoStackId());
		buffer.writeFloat((float)(-profiles[i]->getDepth()));
		buffer.writeFloat((float)profiles[i]->getWaterThick());
	}

	//******************************************************
	//write Tessellation file
	//******************************************************

	string s = "SLBM Tessellation";
	buffer.writeString(s);

	s = "Parameter list";
	buffer.writeString(s);

	s = "Comment";
	buffer.writeString(s);

	buffer.writeInt32(profiles.size());

	for (int i=0; i<profiles.size(); i++)
	{
		buffer.writeDouble(profiles[i]->getLatDegrees());
		buffer.writeDouble(profiles[i]->getLonDegrees());
	}

	buffer.writeInt32(triangles.size());

	for (int i=0; i<triangles.size(); i++)
		for (int j=0; j<3; j++)
			buffer.writeInt32(triangles[i]->getNode(j)->getNodeId());

	buffer.writeInt32(0); // no additional metadata
}

GreatCircle* Grid::getGreatCircle(
		const int& phase,
		const double& latSource, 
		const double& lonSource,
		const double& depthSource,
		const double& latReceiver,
		const double& lonReceiver,
		const double& depthReceiver,
		const double& delta,
		const double& ch_max)
{
	return GreatCircleFactory::create(
		phase, this,
		latSource, lonSource, depthSource,
		latReceiver, lonReceiver, depthReceiver,
		ch_max);
}

void Grid::getNodeNeighbors(const int& nodeid, int neighbors[], int& nNeighbors)
{
	double coeff[nCoefficients];
	Triangle* start = findTriangle(*profiles[nodeid], coeff);
	set<int> nbrs;
	start->findNodeNeighbors(nodeid, nbrs);

	nNeighbors = 0;
	for (set<int>::iterator it=nbrs.begin(); it != nbrs.end(); it++)
			neighbors[nNeighbors++] = profiles[*it]->getNodeId();
}

void Grid::getActiveNodeNeighbors(const int& nodeid, int neighbors[], int& nNeighbors)
{
	double coeff[nCoefficients];
	int id = getGridNodeId(nodeid);
	Triangle* start = findTriangle(*profiles[id], coeff);
	set<int> nbrs;
	start->findNodeNeighbors(id, nbrs);

	nNeighbors = 0;
	for (set<int>::iterator it=nbrs.begin(); it != nbrs.end(); it++)
	{
		id = getActiveNodeId(profiles[*it]->getNodeId());
		if (id >= 0)
			neighbors[nNeighbors++] = id;
	}
}

void Grid::getNodeNeighbors(const int& nodeid, vector<int>& neighbors)
{
	double coeff[nCoefficients];
	Triangle* start = findTriangle(*profiles[nodeid], coeff);
	set<int> nbrs;
	start->findNodeNeighbors(nodeid, nbrs);

	neighbors.clear();
	for (set<int>::iterator it=nbrs.begin(); it != nbrs.end(); it++)
			neighbors.push_back(profiles[*it]->getNodeId());
}

void Grid::getActiveNodeNeighbors(const int& nodeid, vector<int>& neighbors)
{
	double coeff[nCoefficients];
	int id = getGridNodeId(nodeid);
	Triangle* start = findTriangle(*profiles[id], coeff);
	set<int> nbrs;
	start->findNodeNeighbors(id, nbrs);

	neighbors.clear();
	for (set<int>::iterator it=nbrs.begin(); it != nbrs.end(); it++)
	{
		id = getActiveNodeId(profiles[*it]->getNodeId());
		if (id >= 0)
			neighbors.push_back(id);
	}
}

void Grid::getNodeNeighborInfo(const int& nid, int neighbors[], 
			double distance[], double azimuth[], int& nNeighbors)
{
	getNodeNeighbors(nid, neighbors, nNeighbors);
	for (int i=0; i<nNeighbors; i++)
	{
		getNodeSeparation(nid, neighbors[i], distance[i]);
		getNodeAzimuth(nid, neighbors[i], azimuth[i]);
	}
}

void Grid::getActiveNodeNeighborInfo(const int& nid, int neighbors[], 
			double distance[], double azimuth[], int& nNeighbors)
{
	getActiveNodeNeighbors(nid, neighbors, nNeighbors);
	int id = getGridNodeId(nid);
	for (int i=0; i<nNeighbors; i++)
	{
		getNodeSeparation(id, getGridNodeId(neighbors[i]), distance[i]);
		getNodeAzimuth(id, getGridNodeId(neighbors[i]), azimuth[i]);
	}
}

void Grid::getNodeNeighborInfo(const int& nid, vector<int>& neighbors, 
		vector<double>& distance, vector<double>& azimuth)
{
	getNodeNeighbors(nid, neighbors);
	distance.resize(neighbors.size());
	azimuth.resize(neighbors.size());
	for (int i=0; i<neighbors.size(); i++)
	{
		getNodeSeparation(nid, neighbors[i], distance[i]);
		getNodeAzimuth(nid, neighbors[i], azimuth[i]);
	}
}

void Grid::getActiveNodeNeighborInfo(const int& nid, vector<int>& neighbors, 
		vector<double>& distance, vector<double>& azimuth)
{
	getActiveNodeNeighbors(nid, neighbors);
	int id = getGridNodeId(nid);
	distance.resize(neighbors.size());
	azimuth.resize(neighbors.size());
	for (int i=0; i<neighbors.size(); i++)
	{
		getNodeSeparation(id, getGridNodeId(neighbors[i]), distance[i]);
		getNodeAzimuth(id, getGridNodeId(neighbors[i]), azimuth[i]);
	}
}

void Grid::getNodeSeparation(const int& node1, const int& node2, double& distance)
{
	distance = profiles[node1]->distance(*profiles[node2]);
}

void Grid::getNodeAzimuth(const int& node1, const int& node2, double& azimuth)
{
	azimuth = profiles[node1]->azimuth(*profiles[node2]);
}

void Grid::initializeActiveNodes(double activeNodeLatMin, 
								 double activeNodeLonMin, 
								 double activeNodeLatMax, 
								 double activeNodeLonMax) 
{ 
	// ensure that lonMin is >= -PI and < PI
	while (activeNodeLonMin < -PI) activeNodeLonMin += 2*PI;
	while (activeNodeLonMin >= PI) activeNodeLonMin -= 2*PI;
	// ensure that lonMax > lonMin and <= lonMin+2*PI
	while (activeNodeLonMax <=  activeNodeLonMin) activeNodeLonMax += 2*PI;
	while (activeNodeLonMax > activeNodeLonMin+2*PI) activeNodeLonMax -= 2*PI;

	double lat, lon;
	bool active;
	for (int i=0; i<triangles.size(); i++)
	{
		active = false;
		for (int j=0; j<3; j++)
		{
			lon = triangles[i]->getNode(j)->getLon();
			if (lon < activeNodeLonMin) lon += 2*PI;

			if (lon <= activeNodeLonMax)
			{
				lat = triangles[i]->getNode(j)->getLat();
				if (lat >= activeNodeLatMin && lat <= activeNodeLatMax)
				{
					active = true;
					break;
				}
			}
		}
		if (active) 
		for (int j=0; j<3; j++)
		if (triangles[i]->getNode(j)->getActiveNodeId() < 0)
		{
			triangles[i]->getNode(j)->setActiveNodeId(activeNodes.size());
			activeNodes.push_back(triangles[i]->getNode(j)->getNodeId());
		}
	}

	//while (activeNodeLonMin >= PI) activeNodeLonMin -= 2*PI;
	//while (activeNodeLonMin < -PI) activeNodeLonMin += 2*PI;
	//while (activeNodeLonMax >= PI) activeNodeLonMax -= 2*PI;
	//while (activeNodeLonMax < -PI) activeNodeLonMax += 2*PI;
	//double lat, lon;
	//bool active;
	//for (int i=0; i<triangles.size(); i++)
	//{
	//	active = false;
	//	for (int j=0; j<3; j++)
	//	{
	//		lon = triangles[i]->getNode(j)->getLon();
	//		if ((activeNodeLonMin < activeNodeLonMax &&
	//			lon >= activeNodeLonMin && lon <= activeNodeLonMax)
	//			|| (activeNodeLonMin > activeNodeLonMax &&
	//			lon <= activeNodeLonMin && lon >= activeNodeLonMax))
	//		{
	//			lat = triangles[i]->getNode(j)->getLat();
	//			if (lat >= activeNodeLatMin && lat <= activeNodeLatMax)
	//			{
	//				active = true;
	//				break;
	//			}
	//		}
	//	}
	//	if (active) 
	//	for (int j=0; j<3; j++)
	//	if (triangles[i]->getNode(j)->getActiveNodeId() < 0)
	//	{
	//		triangles[i]->getNode(j)->setActiveNodeId(activeNodes.size());
	//		activeNodes.push_back(triangles[i]->getNode(j)->getNodeId());
	//	}
	//}
}

void Grid::writeBufferToFile(util::DataBuffer& buffer, string fileName)
{
	//cout << "Grid::writeBufferToFile() Writing file " << endl << fileName << endl;

	// Open file.
	std::ofstream outFile;
	outFile.open(fileName.c_str(), ios::out | ios::binary);
	if (!outFile.is_open())
	{
		ostringstream os;
		os << endl << "ERROR in Grid::writeBufferToFile()." << endl
			<<"Cannot open file " << fileName << endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),105);
	}

	// Write the contents of the buffer to the file.
	buffer.writeToFile(outFile);

	// Clean up.
	outFile.close();
}

void Grid::reaDataBuffererFromFile(util::DataBuffer& buffer, string dirname, string fileName)
{
	//default to SunOS
	char pathSeparator = '/';
	#if defined(_WIN32) || defined(WIN32)
		pathSeparator = '\\';
	#endif

	if ( dirname.find_last_of(pathSeparator) != dirname.length() - 1 )
		dirname += pathSeparator;

	string pathFileName = dirname + fileName;

	ifstream dataFile(pathFileName.c_str(), ios::in | ios::binary);

	if (dataFile.fail() || !dataFile.is_open())
	{
		ostringstream os;
		os << endl << "ERROR in Grid::reaDataBuffererFromFile" << endl
			<<"Could not open file " << pathFileName << endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ 
			<< " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),106);
	}

	// Get size of file.
	dataFile.seekg(0, ios::end);
	int fileSize = dataFile.tellg();

	// Read entire file into a buffer.
	buffer.readFromFile(dataFile, 0, fileSize);
  
  //reverse byte order on little endian machines
  if(!util::MD50::isBigEndian()) buffer.setByteOrderReverse(true);

	// reset buffer position 
  buffer.resetPos();
	dataFile.close();
}

// **** _FUNCTION DESCRIPTION_ *************************************************
//
//! \brief Returns true if the directory exists. 
//
// *****************************************************************************
bool Grid::is_directory(const string& dir)
{
#if defined(_WIN32) || defined(WIN32)  // Windows 
 
  // check if the directory exists
  if((_access(dir.c_str(), 0)) != -1)  
    return true;
  else 
    return false;

#else  // Sun

  struct dirent *ep;
  struct stat st;

  // read directory information into st
  if (stat(dir.c_str(), &st) != 0)
    return false;

  // return true if it's a directory by checking the mode
  if (S_ISDIR(st.st_mode) != 0) // returns non-zero if it's a directory
    return true;
  else
    return false;

#endif
}

void Grid::findSpecialTriangles()
{
	// specify the unit vectors of 42 positions on the earth that are approximatly
	// evenly distributed on the surface.  These were generated by starting with 12 points 
	// that define an icosahedron. Then add the midpoint of each of the 30 edges.
	int nSpecial = 42;
	double v[42][3];
	int i=0;
	v[i][0] =   0.00000000000000; v[i][1] =   0.00000000000000; v[i++][2] =   1.00000000000000;
	v[i][0] =   0.89442719099992; v[i][1] =  -0.00000000000000; v[i++][2] =   0.44721359549996;
	v[i][0] =   0.27639320225002; v[i][1] =   0.85065080835204; v[i++][2] =   0.44721359549996;
	v[i][0] =  -0.72360679774998; v[i][1] =   0.52573111211913; v[i++][2] =   0.44721359549996;
	v[i][0] =  -0.72360679774998; v[i][1] =  -0.52573111211913; v[i++][2] =   0.44721359549996;
	v[i][0] =   0.27639320225002; v[i][1] =  -0.85065080835204; v[i++][2] =   0.44721359549996;
	v[i][0] =   0.72360679774998; v[i][1] =  -0.52573111211913; v[i++][2] =  -0.44721359549996;
	v[i][0] =   0.72360679774998; v[i][1] =   0.52573111211913; v[i++][2] =  -0.44721359549996;
	v[i][0] =  -0.27639320225002; v[i][1] =   0.85065080835204; v[i++][2] =  -0.44721359549996;
	v[i][0] =  -0.89442719099992; v[i][1] =   0.00000000000000; v[i++][2] =  -0.44721359549996;
	v[i][0] =  -0.27639320225002; v[i][1] =  -0.85065080835204; v[i++][2] =  -0.44721359549996;
	v[i][0] =   0.00000000000000; v[i][1] =   0.00000000000000; v[i++][2] =  -1.00000000000000;
	v[i][0] =   0.16245984811645; v[i][1] =   0.50000000000000; v[i++][2] =   0.85065080835204;
	v[i][0] =   0.68819096023559; v[i][1] =   0.50000000000000; v[i++][2] =   0.52573111211913;
	v[i][0] =   0.52573111211913; v[i][1] =   0.00000000000000; v[i++][2] =   0.85065080835204;
	v[i][0] =  -0.42532540417602; v[i][1] =   0.30901699437495; v[i++][2] =   0.85065080835204;
	v[i][0] =  -0.26286555605957; v[i][1] =   0.80901699437495; v[i++][2] =   0.52573111211913;
	v[i][0] =  -0.42532540417602; v[i][1] =  -0.30901699437495; v[i++][2] =   0.85065080835204;
	v[i][0] =  -0.85065080835204; v[i][1] =   0.00000000000000; v[i++][2] =   0.52573111211913;
	v[i][0] =   0.16245984811645; v[i][1] =  -0.50000000000000; v[i++][2] =   0.85065080835204;
	v[i][0] =  -0.26286555605957; v[i][1] =  -0.80901699437495; v[i++][2] =   0.52573111211913;
	v[i][0] =   0.68819096023559; v[i][1] =  -0.50000000000000; v[i++][2] =   0.52573111211913;
	v[i][0] =   0.58778525229247; v[i][1] =   0.80901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =   0.95105651629515; v[i][1] =   0.30901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =  -0.58778525229247; v[i][1] =   0.80901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =   0.00000000000000; v[i][1] =   1.00000000000000; v[i++][2] =   0.00000000000000;
	v[i][0] =  -0.95105651629515; v[i][1] =  -0.30901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =  -0.95105651629515; v[i][1] =   0.30901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =   0.00000000000000; v[i][1] =  -1.00000000000000; v[i++][2] =   0.00000000000000;
	v[i][0] =  -0.58778525229247; v[i][1] =  -0.80901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =   0.95105651629515; v[i][1] =  -0.30901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =   0.58778525229247; v[i][1] =  -0.80901699437495; v[i++][2] =   0.00000000000000;
	v[i][0] =   0.85065080835204; v[i][1] =   0.00000000000000; v[i++][2] =  -0.52573111211913;
	v[i][0] =   0.26286555605957; v[i][1] =   0.80901699437495; v[i++][2] =  -0.52573111211913;
	v[i][0] =  -0.68819096023559; v[i][1] =   0.50000000000000; v[i++][2] =  -0.52573111211913;
	v[i][0] =  -0.68819096023559; v[i][1] =  -0.50000000000000; v[i++][2] =  -0.52573111211913;
	v[i][0] =   0.26286555605957; v[i][1] =  -0.80901699437495; v[i++][2] =  -0.52573111211913;
	v[i][0] =   0.42532540417602; v[i][1] =   0.30901699437495; v[i++][2] =  -0.85065080835204;
	v[i][0] =   0.42532540417602; v[i][1] =  -0.30901699437495; v[i++][2] =  -0.85065080835204;
	v[i][0] =  -0.16245984811645; v[i][1] =   0.50000000000000; v[i++][2] =  -0.85065080835204;
	v[i][0] =  -0.52573111211913; v[i][1] =   0.00000000000000; v[i++][2] =  -0.85065080835204;
	v[i][0] =  -0.16245984811645; v[i][1] =  -0.50000000000000; v[i++][2] =  -0.85065080835204;

	// define the special triangles.  after a successful triangle walk,
    // the first special triangle will be set to the triangle just found
	// so that the next triangle search will check that triangle first.  
	// In addition, find the triangles which contain the 42 nodes
	// defined above.  It is assumed that these nodes are 
	// evenly distributed over the surface of the earth.  Those triangles
	// will be saved as special triangles.  When Grid::findTriangle() is
	// called, the triangles in special triangles will be searched, in order,
	// for one whose node[0] is within 16 degrees of the location being searched
	// for.  If one is found, then the search will start there.  If none is found,
	// the walk will start at the closest one.
	cos_min = cos(16*DEG_TO_RAD);
	double coefficients[nCoefficients];
	specialTriangles.reserve(nSpecial+1);
	specialTriangles.push_back(triangles[0]);
	Location node;
	for (int i=0; i<nSpecial; i++)
	{
		node.setUnitVector(v[i]);
		specialTriangles.push_back(findTriangle(node, coefficients));
	}
}

} // end slbm namespace
