//- ****************************************************************************
//- 
//- Copyright 2009 Sandia Corporation. Under the terms of Contract 
//- DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains 
//- certain rights in this software.
//-
//- BSD Open Source License.
//- All rights reserved.
//- 
//- Redistribution and use in source and binary forms, with or without 
//- modification, are permitted provided that the following conditions are met:
//-
//-    * Redistributions of source code must retain the above copyright notice, 
//-      this list of conditions and the following disclaimer.
//-    * Redistributions in binary form must reproduce the above copyright 
//-      notice, this list of conditions and the following disclaimer in the 
//-      documentation and/or other materials provided with the distribution.
//-    * Neither the name of Sandia National Laboratories nor the names of its 
//-      contributors may be used to endorse or promote products derived from  
//-      this software without specific prior written permission.
//-
//- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//- ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//- POSSIBILITY OF SUCH DAMAGE.
//-


//- ****************************************************************************
//-
//- Program:       Uncertainty
//- Module:        $RCSfile: Uncertainty.cc,v $
//- Revision:      $Revision: 1.21 $
//- Last Modified: $Date: 2011/10/07 13:15:01 $
//- Last Check-in: $Author: sballar $
//-
//- ****************************************************************************

// **** _SYSTEM INCLUDES_ ******************************************************
#include <fstream>
#include <sstream>
#include <vector>
#include <iostream>

//using namespace std;

#include "Uncertainty.h"
#include "SLBMException.h"

// **** _BEGIN SLBM NAMESPACE_ **************************************************

namespace slbm {

// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Default Constructor
//
// *****************************************************************************
Uncertainty::Uncertainty() 
{
	phaseNum = -1;
	phaseStr = "";
}  // END Uncertainty Default Constructor


// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Parameterized Uncertainty Constructor
//
// *****************************************************************************
Uncertainty::Uncertainty(string modelPath, const string& phase, int phasenum)
{
	phaseStr = phase;
	phaseNum = phasenum;
  fname = modelPath + "/Uncertainty_" + phaseStr + ".txt";
	readFile(fname);
}

// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Parameterized Uncertainty Constructor used by SLBM.  attribute is one of 
// "TT", "Sh", "Az"
//
// *****************************************************************************
Uncertainty::Uncertainty(string modelPath, int phase, int attribute)
{
	switch(phase)
	{
		case 0:
				phaseStr = "Pn";
				break;
		case 1:
				phaseStr = "Sn";
				break;
		case 2:
				phaseStr = "Pg";
				break;
		case 3:
				phaseStr = "Lg";
				break;
		default:
				phaseStr = "Pn";
				break;
	}

	switch(attribute)
	{
		case TT:
				attributeStr = "TT";
				break;
		case SH:
				attributeStr = "Sh";
				break;
		case AZ:
				attributeStr = "Az";
				break;
		default:
				attributeStr = "TT";
				break;
	}

  phaseNum = phase;
  attributeNum = attribute;
  fname = modelPath + "/Uncertainty_" + phaseStr + "_" + attributeStr + ".txt";
  readFile(fname);
}
	
// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Uncertainty Constructor that builds the object from the input DataBuffer.
//
// *****************************************************************************
Uncertainty::Uncertainty(util::DataBuffer& buffer)
{
  deserialize(buffer);
}
	
// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Copy constructor.
//
// *****************************************************************************
Uncertainty::Uncertainty(const Uncertainty& u) :
             phaseNum(u.phaseNum), phaseStr(u.phaseStr),
             fname(u.fname), errDistances(u.errDistances),
             errVal(u.errVal)
{
}

// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Uncertainty Destructor
//
// *****************************************************************************
Uncertainty::~Uncertainty()
{
	phaseNum = -1;
	phaseStr = "";
}
	
// **** _FUNCTION DESCRIPTION_ *************************************************
//
// Assignment operator.
//
// *****************************************************************************
Uncertainty& Uncertainty::operator=(const Uncertainty& u)
{
  phaseNum     = u.phaseNum;
  phaseStr     = u.phaseStr;
  fname        = u.fname;
  errDistances = u.errDistances;
  errVal       = u.errVal;

  return *this;
}

void Uncertainty::readFile(const string& filename)
{
	ifstream fin;
	errDistances.clear();
	errVal.clear();

	//open the file.  Return an error if cannot open the file.
	fin.open(filename.c_str());
	if (fin.fail() || !fin.is_open())
	{
		ostringstream os;
		os << endl << "ERROR in Uncertainty::readFile" << endl
			<<"Could not open file " << filename << endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),115);
	}

	int numdistances = -1;
  int numdepths = 1;

  string buf;
  buf.resize(128);
  fin.getline(&buf[0], buf.size());
  sscanf(&buf[0], "%d %d", &numdistances, &numdepths);
	//fin >> numdistances >> uDepths;

	try
	{
		double temp;

    // read distances

    errDistances.reserve(numdistances);
		for( int i = 0; i < numdistances; i++ )
		{
			fin >> temp;
			if( fin.rdstate() & std::ios::failbit )
			{
				throw SLBMException("",115);
			}
			errDistances.push_back(temp);
		}

    // read depths if > 1

    if (numdepths > 1)
    {
      errDepths.reserve(numdepths);
		  for( int i = 0; i < numdepths; i++ )
		  {
			  fin >> temp;
			  if( fin.rdstate() & std::ios::failbit )
			  {
				  throw SLBMException("",115);
			  }
			  errDepths.push_back(temp);
		  }
    }

    // read errors for each distance and depth

    errVal.resize(numdepths);
    for (int j = 0; j < numdepths; ++j)
    {
		  string tempstr;
		  fin >> tempstr;	//discard # separator line

          errVal[j].reserve(numdistances);
		  for( int i = 0; i < numdistances; i++ )
		  {
			  fin >> temp;
			  if( fin.rdstate() & std::ios::failbit )
			  {
				  throw SLBMException("",115);
			  }

			  // convert slowness from sec/deg to sec/rad
			  if (attributeNum == SH)
				  temp /= DEG_TO_RAD;
			  else if (attributeNum == AZ)
				  temp *= DEG_TO_RAD;

			  errVal[j].push_back(temp);
		  }
    }
	}
	catch( ... )
	{
		ostringstream os;
		os << endl << "ERROR in Uncertainty::readFile" << endl
			<<"Invalid or corrupt file format " << filename << endl
			<< "Version " << SlbmVersion << "  File " << __FILE__ << " line " << __LINE__ << endl << endl;
		throw SLBMException(os.str(),115);
	}
	fin.close();
}

int Uncertainty::getBufferSize() const
{
  int buffsize = 8 * sizeof(int) + (int) phaseStr.size() +
                 (int) fname.size();
  buffsize += (errDistances.size() + 1) * sizeof(double);
  if (errVal.size() > 1)
  {
    buffsize += (errDepths.size()) * sizeof(double);
    buffsize += (errDepths.size() * errDistances.size()) * sizeof(double);
  }
  else
    buffsize += (errDistances.size()) * sizeof(double);

  return buffsize;
}

void Uncertainty::serialize(util::DataBuffer& buffer)
{
  int i;

  // write phaseNum, phaseStr, and fname
  buffer.writeInt32(phaseNum);
  buffer.writeString(phaseStr);
  buffer.writeString(fname);

  // write array distance and depth storage size
  buffer.writeInt32((int) errDistances.size());
  if (errVal.size() == 1)
    buffer.writeInt32(0);
  else
    buffer.writeInt32((int) errDepths.size());

  // Write out vector of distances, depths and uncertainties
  buffer.writeDoubleArray(&errDistances[0], errDistances.size());
  if (errVal.size() > 1)
    buffer.writeDoubleArray(&errDepths[0], errDepths.size());

  // Write out all uncertainties
  for (i = 0; i < errVal.size(); ++i)
    buffer.writeDoubleArray(&errVal[i][0], errDistances.size());  
}

void Uncertainty::deserialize(util::DataBuffer& buffer)
{
  int i;

  // read phaseNum, phaseStr, and fname
  phaseNum = buffer.readInt32();
  phaseStr = buffer.readString();
  fname    = buffer.readString();

  // read distance and depth array storage size
  int ndist = buffer.readInt32();
  int ndpth = buffer.readInt32();

  // resize errDistances and errDepths and read arrays
  errDistances.resize(ndist);
  buffer.readDoubleArray(&errDistances[0], ndist);
  if (ndpth > 0)
  {
    errDepths.resize(ndpth);
    buffer.readDoubleArray(&errDepths[0], ndpth);
  }
  else
  {
    ndpth = 1;
    errDepths.clear();
  }

  // resize errVal and read array
  errVal.resize(ndpth);
  for (i = 0; i < ndpth; ++i)
  {
    errVal[i].resize(ndist);
    buffer.readDoubleArray(&errVal[i][0], ndist);
  }
}

double Uncertainty::getUncertainty(const double& distance, double depth)
{ 
	// Convert to degrees since model errors are defined in degrees.
	double distanceDeg = distance * RAD_TO_DEG;

  // initialize indices and interpolation weights
  int idist, idepth;
  idist = idepth = 0;
  double wdist, wdepth;
  wdist = wdepth = 0.0;

  if ((errVal.size() == 1) || (depth >= errDepths.back()))
  {
    // if more than one depth and depth exceeds last entry set idepth
    if (errVal.size() > 1) idepth = (int) errDepths.size() - 1;

    // if distance in question is greater than the max defined distance, return
    // uncertainty defined for max distance.  Otherwise, interpolate between the
    // two bracketing distances.
	  if( distanceDeg >= errDistances.back())
	    return errVal[idepth].back();
    else
    {
      // get distance interpolation index and weight
      getIndex(distanceDeg, errDistances, idist, wdist);
      return getUncertainty(wdist, idist, idepth);
    }
  }
  else
  {
    // get depth interpolation index and weight
    getIndex(depth, errDepths, idepth, wdepth);

    // if distance in question is greater than the max defined distance, return
    // uncertainty defined for max distance.  Otherwise, interpolate between the
    // two bracketing distances.
	  if( distanceDeg >= errDistances.back())
	    return wdepth * (errVal[idepth+1].back() - errVal[idepth].back()) +
             errVal[idepth].back();
    else
    {
      // get distance interpolation index and weight
      getIndex(distanceDeg, errDistances, idist, wdist);
      return wdepth * (getUncertainty(wdist, idist, idepth+1) - 
                      getUncertainty(wdist, idist, idepth)) +
                      getUncertainty(wdist, idist, idepth);
    }
  }
}

double Uncertainty::getVariance(const double& distance, double depth) 
{ 
	// Convert to degrees since model errors are defined in degrees.
	double distanceDeg = distance * RAD_TO_DEG;

  // initialize indices and interpolation weights
  int idist, idepth;
  idist = idepth = 0;
  double wdist, wdepth;
  wdist = wdepth = 0.0;

  if ((errVal.size() == 1) || (depth >= errDepths.back()))
  {
    // if more than one depth and depth exceeds last entry set idepth
    if (errVal.size() > 1) idepth = (int) errDepths.size() - 1;

    // if distance in question is greater than the max defined distance, return
    // uncertainty defined for max distance.  Otherwise, interpolate between the
    // two bracketing distances.
	  if( distanceDeg >= errDistances.back())
	    return errVal[idepth].back();
    else
    {
      // get distance interpolation index and weight
      getIndex(distanceDeg, errDistances, idist, wdist);
      return getVariance(wdist, idist, idepth);
    }
  }
  else
  {
    // get depth interpolation index and weight
    getIndex(depth, errDepths, idepth, wdepth);

    // if distance in question is greater than the max defined distance, return
    // uncertainty defined for max distance.  Otherwise, interpolate between the
    // two bracketing distances.
	  if( distanceDeg >= errDistances.back())
	    return wdepth * (errVal[idepth+1].back() - errVal[idepth].back()) +
             errVal[idepth].back();
    else
    {
      // get distance interpolation index and weight
      getIndex(distanceDeg, errDistances, idist, wdist);
      return wdepth * (getVariance(wdist, idist, idepth+1) - 
                      getVariance(wdist, idist, idepth)) +
                      getVariance(wdist, idist, idepth);
    }
  }
}

void Uncertainty::getIndex(double x, const vector<double>& v,
                           int& index, double& w) 
{
  // if v consists of only 2 entries then set index to 0 ... otherwise,
  // find bracketing index value
  if (v.size() == 2)
    index = 0;
  else
  {
    // set up interpolation increment and index start
    int inc = (int) v.size();
    inc >>= 1;
    index = inc;

    // perform binary search to find index.
    do
    {
      if (inc > 1) inc >>= 1;
      if (v[index + 1] <= x)
        index += inc;
      else if (v[index] > x)
        index -= inc;
      else
        break;
    } while (true);
  }

  // calculate weight fraction
  w = (x - v[index]) / (v[index+1] - v[index]);
}

} // end slbm namespace
